/* --------------------------------------------------------------------  */
/*                          CALCULIX                                     */
/*                   - GRAPHICAL INTERFACE -                             */
/*                                                                       */
/*     A 3-dimensional pre- and post-processor for finite elements       */
/*              Copyright (C) 1996 Klaus Wittig                          */
/*                                                                       */
/*     This program is free software; you can redistribute it and/or     */
/*     modify it under the terms of the GNU General Public License as    */
/*     published by the Free Software Foundation; version 2 of           */
/*     the License.                                                      */
/*                                                                       */
/*     This program is distributed in the hope that it will be useful,   */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */ 
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the      */
/*     GNU General Public License for more details.                      */
/*                                                                       */
/*     You should have received a copy of the GNU General Public License */
/*     along with this program; if not, write to the Free Software       */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */
/* --------------------------------------------------------------------  */
#include <extUtil.h>



double repairNurbs( Nurbs *nurbs, int nr, int deg, int dir){ return(-1); }
int evalNurbs( Nurbs *nurbs, int nr, int sum_p, double *pnt_u, double *pnt_v, Points *pnt){ return(-1); }
void projSurfToNurbs( Nurbs *nurbs, int nr, Gsur *surf, int snr, Nodes *node ){ return; }
void projPntsToNurbs( Nurbs *nurbs, int nr, Sets *set, int setNr, Points *pnt){ return; }
int getNurbsType( Nurbs *nurbs, int nr, int *axis){ return(-1); }
void trimNurbs( Nurbs *nurbs, int nr, int patch, double TOL_AMBIG){ return(-1); }
