#include <extUtil.h>

int strfind (unsigned char *as1, unsigned char *as2)
{
  int i;
  for(i=0; i<strlen(as1); i++)
  {
    if(as1[i]==as2[0]) if(compare(&as1[i], as2, strlen(as2))==strlen(as2)) return(i);
  }
  return(-1);
}
