/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
#include <extUtil.h>


/* liefert double aus string von position a bis b */
double stof(unsigned char *string, int a, int b)
{
  register int    n, i;
  static unsigned char  puffer[MAX_LINE_LENGTH];

  n=-1;
  for (i=a-1; i<b; i++)
    {
    n++;
    if ((i>=MAX_LINE_LENGTH)||(n>=MAX_LINE_LENGTH)) break;
    puffer[n] = string[i];
  }
  puffer[n+1] = (int)NULL;
  return( atof( puffer ) );
}

