// libSNL - Simple Nurbs Library
// Copyright Scott A.E. Lanham, Australia.
// ---------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU Library General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef SNLCTRLPOINT_H
#define SNLCTRLPOINT_H

#include "snlPoint.h"

class snlCtrlPoint : public snlPoint
{
    public:

        snlCtrlPoint();        
        
        snlCtrlPoint ( snlPoint& );
        
        void operator = ( const snlPoint& copyFrom );

        void select ( bool yesNo );
        bool isSelected();
        
        void weight ( double setTo );
        double weight();

    private:

        bool    selected;  // Point has been selected.
};

#endif
